/* Parallel port emulated EPROM loader for
 * Theo Markettos' Home Automation System
 *
 * This has only been tested on an Acorn Risc PC
 * using Acorn's ANSI C release 4
 *
 * Most of it is ANSI C, although there are a few
 * Risc OS specific parts, such as opening the
 * printer device, which can been removed for other
 * OS's
 *
 * It uses a standard bidirectional parallel port,
 * accessed via the parallel_ calls.  If you wish
 * to use this code, you will need to implement these
 * calls for your hardware
 *
 * This code is copyright Theo Markettos 1997,
 * and may be freely distributed, with the proviso
 * that it is used for educational and non-profit
 * making purposes only
 *
 * If you have any problems using this code, you
 * should contact the author at
 * theomarkettos@letterbox.com
 *
 */

/* ANSI includes */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
/* OSLib includes */
#include "osfind.h"
#include "parallel.h"
#include "osbyte.h"
/* ATMsLib includes - ATMsLib must be before OSLib in include path */
#include "fileutils.h"
#include "parallel.h"
#include "atmstring.h"
#include "atmcore.h" /* Borrowed from DeskLib */

/* ROM base address in terms of printer port interface */
#define ROMBASE 0x000000
/* ROM size in kilobytes */
#define ROMSIZE 16

/*#define DEBUGON*/

os_f printer_device;

void initlink(void);
void resetlink(void);
void romload(FILE *file,long start,long length);
int srecload(FILE *file,long start,long maxlen);
void write_word(long adr,long value);
void set_adr(long adr);
void write_data(long word);
long read_data(void);
void delay(void);
int hextodec(char *string,int ptr);
void uppercase(char *str);
long int fextent(FILE *file);

int main(int argc,char *argv[])
{
  FILE *infile;
  BOOL srec=FALSE,bin=FALSE;
  int loop;
  long size;
  char filename[256];
  char fName[640], outName[640], temp[640], *p, *q, *r, s;
  int i,j=0;

  printf("Emulated EPROM loader by ATM\n");
  printf("Version 1.08 4/4/97\n\n");

  if ((argc<2) || (argc>3))
  {
    fprintf(stderr,"Syntax: %s [-b <binary file>]|[-s <S-record file>]\n",argv[0]);
    return(1);
  }

  for (loop=1;loop<argc;loop++)
    uppercase(argv[loop]);

  if (strcmp(argv[1],"-B")==0)
  {
    bin=TRUE;
    strcpy(filename,argv[2]);
  }
  else
  {
    if (strcmp(argv[1],"-S")==0)
    {
      srec=TRUE;
      strcpy(filename,argv[2]);
    }
    else
    {
      srec=TRUE;
      strcpy(filename,argv[1]);
    }
  }

	/* Swap round last two bits of filename */
	strcpy(outName, filename);
	strcpy(fName, filename);
	p=strrchr(fName,'.');
	if (p==0)
	{ j=1;
	  strcpy(temp,"@.");
	  strcat(temp,fName);
	  p=strrchr(temp,'.');
	  strcpy(outName,"@.");
	  strcat(outName,fName);
	  /*printf("temp=%s, outName=%s\n",temp,outName);*/
	}
	q=strrchr(outName,'.');
	if (q==0)
	{ j=1;
	  q=outName;
	}
	*q='\0';
	r=strrchr(outName,'.');
	if (r==0)
	{ j=1;
	  strcpy(temp,"@.");
	  strcat(temp,fName);
	  p=strrchr(temp,'.');
	  strcpy(outName,"@.");
	  strcat(outName,fName);
	  q=strrchr(outName,'.');
          *q='\0';
          r=strrchr(outName,'.');

	  /*printf("temp=%s, outName=%s\n",temp,outName);*/

	}
	strcpy(temp,r);
	strcpy(r,p); /* Copy last bit of out to 2nd last of file */
	/*strcat(outName,".");*/
	strncat(outName,temp,(q-r));
	/*printf("Reading file %s\n",outName);*/
	strcpy(filename,outName);




/*  switch (argv[1])
  {
    case "-B":
      bin=TRUE;
      strcpy(filename,argv[2]);
      break;
    case "-S":
      srec=TRUE;
      strcpy(filename,argv[2]);
      break;
    default:
      srec=TRUE;
      strcpy(filename,argv[1]);
      break;
  } */


  if (!(infile=fopen(filename,"r")))
  {
    fprintf(stderr,"File '%s' not found\n",argv[1]);
    return(1);
  }

  initlink();
  size=fextent(infile);
  if (bin)
  {
    if (size>ROMSIZE*1024)
    {
      size=ROMSIZE*1024;
      fprintf(stderr,"Warning: File is greater than installed RAM. Only storing first %ld bytes\n",size);
    }
    printf("Uploading binary file...");
    romload(infile,ROMBASE,size);
  }
  if (srec)
  {
    printf("Uploading Motorola S-record...");
    srecload(infile,ROMBASE,ROMSIZE*1024);
  }
  putchar('\n');

  resetlink();

  fclose(infile);
  printf("File '%s' loaded successfully\n",filename);
  return 0;
}

void initlink(void)
{
  printer_device=osfind_openin(0,"parallel:","");
  parallel_setinput();
}

void resetlink(void)
{
  parallel_setinput();
  osfind_close(printer_device);
}

void romload(FILE *file,long start,long length)
{
  long ptr,word;
  char byteh,bytel;

  for (ptr=start;ptr<start+length;ptr+=2)
  {
    byteh=fgetc(file);
    bytel=fgetc(file);
    write_word(ptr,(long) ((byteh<<8) | bytel));
    if ((ptr-start) % 0x40)
      putchar('.');
  }
}

int srecload(FILE *file,long start,long maxlen)
{
  char linebuf[256];
  int lineno=0,length,end,checksum,loop,adr;
  long word;
  FILE *outfile;

#ifdef DEBUGON
outfile=fopen("<Wimp$Scrap>","wb");
#endif
  while (!feof(file))
  {
    lineno++;
    fgets(linebuf,256,file);
    if (linebuf[0]!='S')
    {
      fprintf(stderr,"Error at S-record file line %d: Non S-record data found\n",lineno);
      return(1);
    }
    if (isdigit(linebuf[1])==0)
    {
      fprintf(stderr,"Error at S-recond file line %d: Invalid S-record type\n",lineno);
      return(1);
    }
    length=hextodec(linebuf,2);
    end=strlen(linebuf);
    /*checksum=((linebuf[end-2]-'0')<<4) | ((linebuf[end-1]-'0'));*/
    checksum=hextodec(linebuf,length*2+3-1);
#ifdef DEBUGON
    printf("line %d, length %X, end %d,checksum=%X\n",lineno,length,end,checksum);
    printf("line itself: %s\n",linebuf);
#endif
    putchar('.');
    switch (linebuf[1])
    {
      case '0':  /* Header record */
        /* ignore header record */
        break;

      case '1':  /* 2 byte adr, code/data */
        adr=(hextodec(linebuf,4)<<8) | (hextodec(linebuf,6));
        if (adr>maxlen)
          fprintf(stderr,"Warning at S-record line %d: attempt to write to nonexistent memory\n",lineno);
        adr+=start;
	for (loop=8;loop<(length)*2;loop+=4)
	{
	  word=(hextodec(linebuf,loop)<<8) | (hextodec(linebuf,loop+2));
	  write_word(adr+(loop-8)/2,word);
#ifdef DEBUGON
	  fseek(outfile,adr+(loop-8)/2,SEEK_SET);
	  fputc((word & 0xFF00)>>8,outfile);
	  fputc(word & 0xFF,outfile);
	  printf("Written word $%X to $%X\n",word,adr+(loop-8)/2);
#endif
	}
        break;

      case '9':  /* Terminator */
        adr=(hextodec(linebuf,4)<<8) | (hextodec(linebuf,6));
	if (adr!=0)
	  printf("Control would now pass to $%X\n",adr);
	break;
    }
  }
  return(0);
}


void write_word(long adr,long value)
{
  long checkval;
  int i;
  adr=adr/2;
  set_adr(adr);
  write_data(value);
  delay();
  set_adr(0);
  set_adr(adr);
  checkval=read_data();
  for (i=0;i<50;i++)
  {
  if (checkval!=(value & 0xFFFF))
  {
    fprintf(stderr,"*",adr*2);
    delay();
    delay();
    delay();
    delay();
  parallel_setctrl(0xF);
    delay();
  /*printf("po ret %lX\n",*/parallel_out((adr & 0xFF00)>>8);/*);*/
    delay();
  parallel_setctrl(0xD);  /* %11101 sig high adr */
    delay();
  parallel_setctrl(0x9);  /* %11001 write high adr */
    delay();
  parallel_setctrl(0xD);
    delay();
  parallel_setctrl(0xF);  /* %11111 disable ROM */
    delay();
  /*printf("po ret %lX\n",*/parallel_out(adr & 0xFF);/*);*/
    delay();
  parallel_setctrl(0xE);  /* %11110 sig low adr */
    delay();
  parallel_setctrl(0xA);  /* %11010 write low adr */
    delay();
  parallel_setctrl(0xE);
    delay();
  parallel_setctrl(0xF);  /* %11111 disable ROM */
    delay();
  parallel_setinput();



    /*set_adr(adr);*/
    delay();
    delay();
    delay();
    delay();
  parallel_setctrl(0xF);
    delay();
  /*printf("po ret %lX\n",*/parallel_out((value & 0xFF00)>>8);/*);*/
    delay();
  parallel_setctrl(0x4);
    delay();
  parallel_setctrl(0x0); /* %0000 write high data */
    delay();
  parallel_setctrl(0x4);
    delay();
  parallel_out(0x7F);
  parallel_out(0x3F);
  parallel_out(0x1F);
  parallel_out(0x0F);
  parallel_out(0x07);
  parallel_out(0x03);
  parallel_out(0x01);
  parallel_out(0x00);
  parallel_setctrl(0xC);
    delay();
  parallel_setctrl(0xF); /* %11111 disable ROM */
    delay();

/*if (((value & 0xFF00)>>8)==0xFF)
{ */
/*}*/

  parallel_setinput();
    delay();

  parallel_setctrl(0xF);
    delay();
  parallel_setctrl(0xC); /* %11100 sig high data */
    delay();
  parallel_setctrl(0x8); /* %11000 read high data */
    delay();
  checkval=parallel_in()<<8;
    delay();
  parallel_setctrl(0xC);
    delay();
  parallel_setctrl(0xF); /* %11111 disable ROM */
    delay();
  /*if (checkval!=(value & 0xFF00))
    fprintf(stderr,"\nTop byte at $%lX, in %lX, out %lX",adr*2,value & 0xFF00,checkval);*/

  /*printf("po ret %lX\n",*/parallel_out(value & 0xFF);/*);*/
    delay();
  parallel_setctrl(0x7);
    delay();
  parallel_setctrl(0x3); /* %0011 write low data */
    delay();
  parallel_setctrl(0x7);
    delay();
  parallel_setinput();
    delay();
  parallel_setctrl(0xF); /* %11111 disable ROM */
    /*write_data(value);*/
    delay();
    delay();
    delay();
    delay();
  parallel_setctrl(0xF);
    delay();
  /*printf("po ret %lX\n",*/parallel_out(0);/*);*/
    delay();
  parallel_setctrl(0xD);  /* %11101 sig high adr */
    delay();
  parallel_setctrl(0x9);  /* %11001 write high adr */
    delay();
  parallel_setctrl(0xD);
    delay();
  parallel_setctrl(0xF);  /* %11111 disable ROM */
    delay();
  /*printf("po ret %lX\n",*/parallel_out(0);/*);*/
    delay();
  parallel_setctrl(0xE);  /* %11110 sig low adr */
    delay();
  parallel_setctrl(0xA);  /* %11010 write low adr */
    delay();
  parallel_setctrl(0xE);
    delay();
  parallel_setctrl(0xF);  /* %11111 disable ROM */
    delay();
  parallel_setinput();
    /*set_adr(0);*/
    delay();
    delay();
    delay();
    delay();
  parallel_setctrl(0xF);
    delay();
  /*printf("po ret %lX\n",*/parallel_out((adr & 0xFF00)>>8);/*)*/
    delay();
  parallel_setctrl(0xD);  /* %11101 sig high adr */
    delay();
  parallel_setctrl(0x9);  /* %11001 write high adr */
    delay();
  parallel_setctrl(0xD);
    delay();
  parallel_setctrl(0xF);  /* %11111 disable ROM */
    delay();
  /*printf("po ret %lX\n",*/parallel_out(adr & 0xFF);/*);*/
    delay();
  parallel_setctrl(0xE);  /* %11110 sig low adr */
    delay();
  parallel_setctrl(0xA);  /* %11010 write low adr */
    delay();
  parallel_setctrl(0xE);
    delay();
  parallel_setctrl(0xF);  /* %11111 disable ROM */
    delay();
  parallel_setinput();
    /*set_adr(adr);*/
    delay();
    delay();
    delay();
    delay();
  parallel_setctrl(0xF);
    delay();
  parallel_setctrl(0xC); /* %11100 sig high data */
    delay();
  parallel_setctrl(0x8); /* %11000 read high data */
    delay();
  checkval=parallel_in()<<8;
    delay();
  parallel_setctrl(0xC);
    delay();
  parallel_setctrl(0xF); /* %11111 disable ROM */
    delay();
  parallel_setctrl(0xF); /* %11111 sig high data */
    delay();
  parallel_setctrl(0xB); /* %11011 read low data */
    delay();
  checkval|=parallel_in();
    delay();
  parallel_setctrl(0xF); /* %11111 disable ROM */

    /*checkval=read_data();*/
  }
  if (checkval==(value & 0xFFFF))
    break;

  }
    if (checkval!=(value & 0xFFFF))
    {
      fprintf(stderr,"\nMemory error at $%lX, in $%lX, out $%lX\n",adr*2,value,checkval);
      exit(1);
    }

}

void set_adr(long adr)
{
  parallel_setctrl(0xF);
  parallel_out((adr & 0xFF00)>>8);
  parallel_setctrl(0xD);  /* %11101 sig high adr */
  parallel_setctrl(0x9);  /* %11001 write high adr */
  parallel_setctrl(0xD);
  parallel_setctrl(0xF);  /* %11111 disable ROM */
  parallel_out(adr & 0xFF);
  parallel_setctrl(0xE);  /* %11110 sig low adr */
  parallel_setctrl(0xA);  /* %11010 write low adr */
  parallel_setctrl(0xE);
  parallel_setctrl(0xF);  /* %11111 disable ROM */
  parallel_setinput();
}

void write_data(long word)
{
  parallel_setctrl(0xF);
  parallel_out((word & 0xFF00)>>8);
  parallel_setctrl(0x4);
  parallel_setctrl(0x0); /* %0000 write high data */
  parallel_setctrl(0x4);
  parallel_setinput();
  parallel_setctrl(0xC);
  parallel_setctrl(0xF); /* %11111 disable ROM */
  parallel_out(word & 0xFF);
  parallel_setctrl(0x7);
  parallel_setctrl(0x3); /* %0011 write low data */
  parallel_setctrl(0x7);
  parallel_setinput();
  parallel_setctrl(0xF); /* %11111 disable ROM */
}

long read_data(void)
{
  long data;

  delay();

  parallel_setctrl(0xF);
  parallel_setctrl(0xC); /* %11100 sig high data */
  parallel_setctrl(0x8); /* %11000 read high data */
  data=parallel_in()<<8;
  parallel_setctrl(0xC);
  parallel_setctrl(0xF); /* %11111 disable ROM */
  parallel_setctrl(0xF); /* %11111 sig high data */
  parallel_setctrl(0xB); /* %11011 read low data */
  data|=parallel_in();
  parallel_setctrl(0xF); /* %11111 disable ROM */
  return(data);
}

void delay(void)
{
int l;
  /*osbyte(osbyte_AWAIT_VSYNC,0,0);
  osbyte(osbyte_AWAIT_VSYNC,0,0);  */
  for (l=0;l<100000;l++);
}

int hextodec(char *string,int ptr)
{
  char msb,lsb;
  int  hnyb,lnyb;

  msb=string[ptr];
  lsb=string[ptr+1];

  hnyb=msb-'0';
  if (msb>'9')
    hnyb=msb+10-'A';

  lnyb=lsb-'0';
  if (lsb>'9')
    lnyb=lsb+10-'A';

  return ((hnyb<<4)|(lnyb));
}


void uppercase(char *str) /*Replaced by ARM code version*/
{
  int ptr=0;

  while (str[ptr])
  {
    str[ptr]=toupper(str[ptr]);
    ptr++;
  }
}

long int fextent(FILE *file)
{
  fpos_t original_position;
  long int endpos;

  if (fgetpos(file,&original_position)) return(-1);

  if (fseek(file,0,SEEK_END)) return(-1);

  if ((endpos=ftell(file))==-1L) return(-1);

  if (fsetpos(file,&original_position)) return(-1);

  return(endpos);
}
